//	$Header: /afs/fl.mcs.anl.gov/usr/cave/hudson/EECS590/EECS590.FL/VRenClass/RCS/Main.C,v 1.30 1996/06/24 22:07:08 hudson Exp $

//	VRenMain.C is for EECS590 project....


#include "VRen.h"
#include "sharedm.h"				//	Terry's
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <device.h>
#include <iostream.h>
#include <assert.h>
#include <string.h>
#include <GUI_Menu.h>


SharedMemArena *aSharedMemArena;

//	A Dan Heath menu...
tMenu H_Menu;
enum MenuItems {NULL_ITEM,
					 H_MENU_ISOMETRY,
					 H_MENU_SCALE, H_MENU_TRANSLATE, H_MENU_ROTATE,
					 H_MENU_LUT,
					 H_MENU_SHIFT_PROF,
					 H_MENU_WIDTH_PROF,
					 H_MENU_HEIGHT_PROF,
					 H_MENU_RPROF,
					 H_MENU_GPROF,
					 H_MENU_BPROF,
					 H_MENU_APROF,
					 H_MENU_ALLPROFS,
					 H_MENU_NUM_SLICES,
					 NUM_MENU_ITEMS};

//	Volume rendering constructs
VRen *aVolume;
void *data;
ParmStruct parms;

enum MenuLevel1 {NULL_MLEV1, ISOMETRY, LUT, NUM_SLICES, NUM_MLEV1};
enum MenuLevel2 {NULL_MLEV2, ISO_ROTATE, ISO_SCALE, ISO_TRANSLATE, PROF_SHIFT, PROF_WIDTH, PROF_HEIGHT, NUM_MLEV2};
enum MenuLevel3 {NULL_MLEV3, RPROF, GPROF, BPROF, APROF, ALLPROFS, NUM_MLEV3};
MenuLevel1 *mlev1_sw;
MenuLevel2 *mlev2_sw;
MenuLevel3 *mlev3_sw;
GLfloat *RotAngle;				//	Angle about its center that data volume spins when user grabs and spins it
GLfloat *RotAxis;	  		//	Axis about which data volume spins when user grabs and spins it
GLfloat *Scale;
GLfloat *Translate;
GLfloat *AccumMtx;
GLfloat *AccumRotMtx;
GLint *NumPolys;		//	Number of slicing polyons (1-500)
CAVELOCK rotmtx_lock;	//	To lock read/write of AccumRotMtx
CAVELOCK rotparms_lock;	//	To lock read/write of RotAngle and Rotaxis


//	           midpoint
//	            __.__                                                     
//	           /     \                                                  
//	          /       \                                                 
//	         /         \    (pretend these diagonals are part of the cos curve)                                            
//	        /           \                                               
//	       /             \                                              
//	 _____/               \____________________________________
//	 0    |<----->|                                           255
//	        halfwidth
//
enum ProfType {RedProf, GreenProf, BlueProf, AlphaProf, NumProfs};
ProfType *CurrProf;
GLfloat *Rprof;	//	(array of) Red color profiles		values: [0.0, 1.0]; currently, cos(x), -PI <= x <= PI
GLfloat *Gprof;	//	(array of) Green color profiles	values: [0.0, 1.0]; currently, cos(x), -PI <= x <= PI
GLfloat *Bprof;	//	(array of) Blue color profiles  	values: [0.0, 1.0]; currently, cos(x), -PI <= x <= PI
GLfloat *Aprof;	//	(array of) Alpha color profiles	values: [0.0, 1.0]; currently, cos(x), -PI <= x <= PI
GLint *RprofMidpoint;	//	index [0, 255] into color table where cos(x) peeks
GLint *GprofMidpoint;	//	index [0, 255] into color table where cos(x) peeks
GLint *BprofMidpoint;	//	index [0, 255] into color table where cos(x) peeks
GLint *AprofMidpoint;	//	index [0, 255] into color table where cos(x) peeks
GLint *RprofHalfwidth;	//	number of table indices that cos(x) spreads left & right of midpoint
GLint *GprofHalfwidth;	//	number of table indices that cos(x) spreads left & right of midpoint
GLint *BprofHalfwidth;	//	number of table indices that cos(x) spreads left & right of midpoint
GLint *AprofHalfwidth;	//	number of table indices that cos(x) spreads left & right of midpoint
GLfloat *RprofAmpFactor;	//	"amplitude factor".  cos curve height is scaled by AmpFactor, then  AmpFactor is
                           //	added to curve.  Result is curve of all + value whose height is <= 1.0.
GLfloat *GprofAmpFactor;
GLfloat *BprofAmpFactor;
GLfloat *AprofAmpFactor;

CAVE	*aCAVE;

void ReadFile(char *filename, size_t filesize, void *data);
void InitSharedData();
void Draw();
void CalcFrame();
void InitApplication();
void FrameApplication();
void BuildSinLUT(ulong256_t my_lut);
void BuildCosLUT(ulong256_t my_lut);
void BuildJeffsLUT(ulong256_t my_lut);
void BuildJeff2LUT(ulong256_t my_lut);
void BuildProf(GLfloat *prof, GLint profMidpoint, GLint profHalfwidth, GLfloat profAmpFactor);
void BuildLUT(ulong256_t my_lut);
void MenuSetup();
void menuCallBack(int message, MenuItems id, tMenuBase *mb);
void ProcessWand(Wand *aWand);
void CalcRotParms(Wand *aWand);
void crossproduct(float *norm, float *firstvec, float *secondvec);
float dot3d(float *vec0, float *vec1);
void normalize(float *vec, float *unitvec);
GLfloat Degrees(GLfloat radians);
void PrintOGLMtx(GLfloat *tempmtx);
void DrawProfile(GLfloat *prof, ProfType CurrProf);


main(int argc, char **argv) {
	int count;

	count = 0;

	aSharedMemArena = new SharedMemArena(megs(48));	//	48
	aSharedMemArena->on();

   CAVEConfigure(&argc,argv,NULL);
	tGUIBase::Init();
	InitSharedData();
	H_Menu.setCallBack(menuCallBack);
	CAVEInit();
	CAVEInitApplication(InitApplication, 0);
	CAVEFrameFunction(FrameApplication, 0);

#if 1	//	testing at anl
	while (!CAVEgetbutton(CAVE_ESCKEY)) {	//	App Loop			
		CalcFrame();			
		CAVEDisplay(Draw, 0);
	}
	CAVEExit();
#else	//	testing at home
	while((count < 10) && (!CAVEgetbutton(CAVE_ESCKEY))) {	//	Demo Loop
		while ((count < 10) && (!CAVEgetbutton(CAVE_ESCKEY))) {	//	App Loop			
			CalcFrame();			
			CAVEDisplay(Draw, 0);
			count++;
		}
		system("scrsave ./snap.bw 0 320 0 256 -b");
	}
	CAVEExit();
#endif
	return 0;
}

void MenuSetup() {
	tSubMenu *sub_menu_main;
	tSubMenu *sub_menu_move;
	tSubMenu *sub_menu_LUT;
	tSubMenu *sub_menu_shiftprof;
	tSubMenu *sub_menu_widthprof;
	tSubMenu *sub_menu_heightprof;

	sub_menu_main = new tSubMenu ("MAIN MENU");
	sub_menu_move = new tSubMenu ("ISOMETRY");
	sub_menu_LUT = new tSubMenu ("LOOKUP TABLE");
	sub_menu_shiftprof = new tSubMenu ("SHIFT PROFILES");
	sub_menu_widthprof = new tSubMenu ("CHANGE PROFILE WIDTHS");
	sub_menu_heightprof = new tSubMenu ("CHANGE PROFILE HEIGHTS");

	sub_menu_move->addItem ("Scale", H_MENU_SCALE);
	sub_menu_move->addItem ("Translate", H_MENU_TRANSLATE);
	sub_menu_move->addItem ("Rotate", H_MENU_ROTATE);

	sub_menu_LUT->addItem (sub_menu_shiftprof, "SHIFT PROFILES -->", H_MENU_SHIFT_PROF);
	sub_menu_LUT->addItem (sub_menu_widthprof, "CHANGE PROFILE WIDTHS -->", H_MENU_WIDTH_PROF);
	sub_menu_LUT->addItem (sub_menu_heightprof, "CHANGE PROFILE HEIGHTS -->", H_MENU_HEIGHT_PROF);

	sub_menu_shiftprof->addItem ("Red profile", H_MENU_RPROF);
	sub_menu_shiftprof->addItem ("Green profile", H_MENU_GPROF);
	sub_menu_shiftprof->addItem ("Blue profile", H_MENU_BPROF);
	sub_menu_shiftprof->addItem ("Alpha profile", H_MENU_APROF);
	sub_menu_shiftprof->addItem ("All profiles", H_MENU_ALLPROFS);

	sub_menu_widthprof->addItem ("Red profile", H_MENU_RPROF);
	sub_menu_widthprof->addItem ("Green profile", H_MENU_GPROF);
	sub_menu_widthprof->addItem ("Blue profile", H_MENU_BPROF);
	sub_menu_widthprof->addItem ("Alpha profile", H_MENU_APROF);
	sub_menu_widthprof->addItem ("All profiles", H_MENU_ALLPROFS);

	sub_menu_heightprof->addItem ("Red profile", H_MENU_RPROF);
	sub_menu_heightprof->addItem ("Green profile", H_MENU_GPROF);
	sub_menu_heightprof->addItem ("Blue profile", H_MENU_BPROF);
	sub_menu_heightprof->addItem ("Alpha profile", H_MENU_APROF);
	sub_menu_heightprof->addItem ("All profiles", H_MENU_ALLPROFS);

	sub_menu_main->addItem (sub_menu_move, "MOVE MODEL -->", H_MENU_ISOMETRY);
	sub_menu_main->addItem (sub_menu_LUT, "CHANGE LOOKUP TABLE -->", H_MENU_LUT);
	sub_menu_main->addItem ("Change # Slices", H_MENU_NUM_SLICES);

	H_Menu.MakeBase (sub_menu_main);
}

void menuCallBack(int message, MenuItems id, tMenuBase *mb) {
	switch (message) {
	case MENU_OPEN:
		break;
	case MENU_CLOSE:
		H_Menu.setSubMenu((tSubMenu*)mb);
		break;
	case MENU_SUB_OPEN:
		switch (id) {
		case H_MENU_ISOMETRY:
			*mlev1_sw = ISOMETRY;
			H_Menu.setSubMenu(H_MENU_ISOMETRY);
			break;
		case H_MENU_LUT:
			*mlev1_sw = LUT;
			H_Menu.setSubMenu(H_MENU_LUT);
			break;
		case H_MENU_SHIFT_PROF:
			*mlev2_sw = PROF_SHIFT;
			H_Menu.setSubMenu(H_MENU_SHIFT_PROF);
			break;
		case H_MENU_WIDTH_PROF:
			*mlev2_sw = PROF_WIDTH;
			H_Menu.setSubMenu(H_MENU_WIDTH_PROF);
			break;
		case H_MENU_HEIGHT_PROF:
			*mlev2_sw = PROF_HEIGHT;
			H_Menu.setSubMenu(H_MENU_HEIGHT_PROF);
			break;
		}
		break;
	case MENU_SUB_CLOSE:
		break;
	case MENU_ITEM:
		switch (id) {
		case H_MENU_NUM_SLICES:
			*mlev1_sw = NUM_SLICES;
			break;
		case H_MENU_SCALE:
			*mlev2_sw = ISO_SCALE;
			break;
		case H_MENU_TRANSLATE:
			*mlev2_sw = ISO_TRANSLATE;
			break;
		case H_MENU_ROTATE:
			*mlev2_sw = ISO_ROTATE;
			break;
		case H_MENU_RPROF:
			*mlev3_sw = RPROF;
			break;
		case H_MENU_GPROF:
			*mlev3_sw = GPROF;
			break;
		case H_MENU_BPROF:
			*mlev3_sw = BPROF;
			break;
		case H_MENU_APROF:
			*mlev3_sw = APROF;
			break;
		case H_MENU_ALLPROFS:
			*mlev3_sw = ALLPROFS;
			break;
		}
		break;
	}
}

void InitSharedData() {
	int i;
	ulong256_t my_lut;

	MenuSetup();

	aCAVE = new CAVE;
	assert(aCAVE != 0);

	RotAngle = new GLfloat();
	RotAxis = new GLfloat[3];
	Scale = new GLfloat();
	Translate = new GLfloat[3];
	AccumMtx = new GLfloat[16];
	AccumRotMtx = new GLfloat[16];
	mlev1_sw = new MenuLevel1();
	mlev2_sw = new MenuLevel2();
	mlev3_sw = new MenuLevel3();
	NumPolys = new GLint();
	*NumPolys = 75;
	Rprof = new GLfloat[256];
	Gprof = new GLfloat[256];
	Bprof = new GLfloat[256];
	Aprof = new GLfloat[256];
	RprofMidpoint = new GLint();
	GprofMidpoint = new GLint();
	BprofMidpoint = new GLint();
	AprofMidpoint = new GLint();
	RprofHalfwidth = new GLint();
	GprofHalfwidth = new GLint();
	BprofHalfwidth = new GLint();
	AprofHalfwidth = new GLint();
	RprofAmpFactor = new GLfloat();
	GprofAmpFactor = new GLfloat();
	BprofAmpFactor = new GLfloat();
	AprofAmpFactor = new GLfloat();
	*RprofMidpoint = *GprofMidpoint = *BprofMidpoint = *AprofMidpoint = 64;
	*RprofHalfwidth = *GprofHalfwidth = *BprofHalfwidth = *AprofHalfwidth = 96;
	*RprofAmpFactor = *GprofAmpFactor = *BprofAmpFactor = *AprofAmpFactor = 0.5;
	CurrProf = new ProfType();
	*CurrProf = RedProf;
	BuildProf(Rprof, *RprofMidpoint, *RprofHalfwidth, *RprofAmpFactor);
	BuildProf(Gprof, *GprofMidpoint, *GprofHalfwidth, *GprofAmpFactor);
	BuildProf(Bprof, *BprofMidpoint, *BprofHalfwidth, *BprofAmpFactor);
	BuildProf(Aprof, *AprofMidpoint, *AprofHalfwidth, *AprofAmpFactor);
	rotmtx_lock = CAVENewLock();
	rotparms_lock = CAVENewLock();

	parms.major_res = parms.middle_res = parms.minor_res = 64;
	data = (uchar_t *) new uchar_t[parms.major_res*parms.middle_res*parms.minor_res];
	parms.major_extent = parms.middle_extent = parms.minor_extent = 1.0;
	parms.res_factor = 1;
	parms.tex_megs = 4;
	sprintf(parms.elem_type, "uchar");
	sprintf(parms.data_name, "Astro");
	ReadFile("datafile", parms.major_res*parms.middle_res*parms.minor_res, data);

	*Scale = 1.0;
	Translate[0] = Translate[1] = Translate[2] = 0.0;
	RotAxis[0] = RotAxis[2] = 0.0;
	RotAxis[1] = 1.0;
	*RotAngle = 0.0;

	BuildLUT(my_lut);
	aVolume = new VRen(data, &parms, *NumPolys, my_lut);
}

//------------------------------------------------------------------------------
void Draw() {	//	Pass this twice to 'cave_display'

	glClearColor(0.0, 0.0, 0.0, 0.0);  glClearDepth(1.0);  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glPushMatrix();
	glMultMatrixf(AccumMtx);
	aVolume->Draw();
	glPopMatrix();
	DrawProfile(Rprof, RedProf);
	DrawProfile(Gprof, GreenProf);
	DrawProfile(Bprof, BlueProf);
	DrawProfile(Aprof, AlphaProf);
	glPushMatrix();
	glTranslatef(0.0, -2.0, -6.0);
	tGUIBase::Draw();
	glPopMatrix();
}

//------------------------------------------------------------------------------
void CalcFrame() {
	ulong256_t my_lut;

	Wand &aWand = aCAVE->GetWand();
//	Head &aHead = aCAVE->GetHead();

	aWand = aCAVE->GetWand();
	ProcessWand(&aWand);

	if (CAVEgetbutton(CAVE_IKEY)) {
		BuildSinLUT(my_lut);
		aVolume->LoadUserLUT(my_lut);
	}
	if (CAVEgetbutton(CAVE_JKEY)) {
		BuildJeff2LUT(my_lut);
		aVolume->LoadUserLUT(my_lut);
	}
	if (CAVEgetbutton(CAVE_OKEY)) {
		BuildCosLUT(my_lut);
		aVolume->LoadUserLUT(my_lut);
	}
	aVolume->CalcFrame();
	aWand.NudgeLoc();
}

//------------------------------------------------------------------------------
void InitApplication() {
	tSubMenu *sub_menu_main;
	tSubMenu *sub_menu_move;

	//	Init AccumMtx
	glPushMatrix();
	glLoadIdentity();
	glGetFloatv(GL_MODELVIEW_MATRIX, AccumMtx);
	glPopMatrix();
	//	Init AccumRotMtx
	CAVESetWriteLock(rotmtx_lock);
	glPushMatrix();
	glLoadIdentity();
	glGetFloatv(GL_MODELVIEW_MATRIX, AccumRotMtx);
	glPopMatrix();
	CAVEUnsetWriteLock(rotmtx_lock);
	
	aVolume->InitApplication();
	tGUIBase::AppInit();
}

void FrameApplication() {
	int i;
	float mtx[16];
	static int modnum = 25;
	int num_polys;
	static GLboolean KEYY_was_off = 1;

	if (CAVEMasterDisplay()) {

		//	Update AccumRotMtx
		CAVESetWriteLock(rotmtx_lock);
		glPushMatrix();
		glLoadIdentity();
		CAVESetReadLock(rotparms_lock);
		glRotatef(*RotAngle, RotAxis[0], RotAxis[1], RotAxis[2]);
		CAVEUnsetReadLock(rotparms_lock);
		glMultMatrixf(AccumRotMtx);
		glGetFloatv(GL_MODELVIEW_MATRIX, AccumRotMtx);
		glPopMatrix();
		CAVEUnsetWriteLock(rotmtx_lock);

		//	Set AccumMtx
		glPushMatrix();
		glLoadIdentity();
		glTranslatef(Translate[0], Translate[1], Translate[2]);
		glScalef(*Scale, *Scale, *Scale);
//		glRotatef(*RotAngle, RotAxis[0], RotAxis[1], RotAxis[2]);
		CAVESetReadLock(rotmtx_lock);
		glMultMatrixf(AccumRotMtx);
		CAVEUnsetReadLock(rotmtx_lock);
		glGetFloatv(GL_MODELVIEW_MATRIX, AccumMtx);
		glPopMatrix();
//		PrintOGLMtx(AccumMtx);
	}
	CAVEDisplayBarrier();

#if 0
	if (KEYY_was_off) {
		if (CAVEgetbutton(CAVE_YKEY)) {
			if (CAVEMasterDisplay()) {
				modnum = (modnum + 75)%550;
//				num_polys = modnum + 50;
				num_polys = modnum;
				KEYY_was_off = 0;
			}
			CAVEDisplayBarrier();
			aVolume->SetNumPolys(num_polys);
		}
	}
	if (!CAVEgetbutton(CAVE_YKEY))
		KEYY_was_off = 1;
#else
	if (*mlev1_sw == NUM_SLICES) {
		CAVEDisplayBarrier();
		aVolume->SetNumPolys(*NumPolys);
	}
#endif

	CAVEDisplayBarrier();		//	Probably should remove this or put it inside VRen::FrameApplication
	//	mtx is the single, net transformation that gets the volume from being
	//	unscaled, centered at the origin and aligned || the 3 axes, to the scale, location
	//	and orientation given it by the sum total of all the user's shenanigans
	aVolume->FrameApplication(AccumMtx);
//	aVolume->FrameApplication(mtx);
	tGUIBase::Frame();
}

void ProcessWand(Wand *aWand) {
	static GLboolean RHB1_was_off = 1;	//	`1' ==> BUTTON1's been OFF since 'twas ON
	static GLboolean RHB2_was_off = 1;	//	`1' ==> BUTTON2's been OFF since 'twas ON
	static GLboolean RHB3_was_off = 1;	//	`1' ==> BUTTON3's been OFF since 'twas ON
	ulong256_t my_lut;

	if (aWand->RefButton2()) {
		switch (*mlev1_sw) {
		case NULL_MLEV1:
			break;
		case ISOMETRY:
			switch (*mlev2_sw) {
			case NULL_MLEV2:
				break;
			case ISO_SCALE:
				*Scale += (aWand->RefLoc(0) - aWand->RefOldloc(0));
				if (*Scale <= 0.0)
					*Scale = 0.1;
				break;
			case ISO_TRANSLATE:
				Translate[0] += ((aWand->RefLoc(0) - aWand->RefOldloc(0)) * 2.0);
				Translate[1] += ((aWand->RefLoc(1) - aWand->RefOldloc(1)) * 2.0);
				Translate[2] += ((aWand->RefLoc(2) - aWand->RefOldloc(2)) * 2.0);
				break;
			case ISO_ROTATE:
				CAVESetWriteLock(rotparms_lock);
				CalcRotParms(aWand);
				CAVEUnsetWriteLock(rotparms_lock);
				break;
			}
			break;
		case LUT:
			switch (*mlev2_sw) {
			case PROF_SHIFT:
				switch (*mlev3_sw) {
				case RPROF:
					*RprofMidpoint += (GLint)(25.0*(aWand->RefLoc(0) - aWand->RefOldloc(0)));
					if (*RprofMidpoint < 0)
						*RprofMidpoint = 0;
					if (*RprofMidpoint > 255)
						*RprofMidpoint = 255;
					BuildProf(Rprof, *RprofMidpoint, *RprofHalfwidth, *RprofAmpFactor);
					BuildLUT(my_lut);
					aVolume->LoadUserLUT(my_lut);
					break;
				case GPROF:
					*GprofMidpoint += (GLint)(25.0*(aWand->RefLoc(0) - aWand->RefOldloc(0)));
					if (*GprofMidpoint < 0)
						*GprofMidpoint = 0;
					if (*GprofMidpoint > 255)
						*GprofMidpoint = 255;
					BuildProf(Gprof, *GprofMidpoint, *GprofHalfwidth, *GprofAmpFactor);
					BuildLUT(my_lut);
					aVolume->LoadUserLUT(my_lut);
					break;
				case BPROF:
					*BprofMidpoint += (GLint)(25.0*(aWand->RefLoc(0) - aWand->RefOldloc(0)));
					if (*BprofMidpoint < 0)
						*BprofMidpoint = 0;
					if (*BprofMidpoint > 255)
						*BprofMidpoint = 255;
					BuildProf(Bprof, *BprofMidpoint, *BprofHalfwidth, *BprofAmpFactor);
					BuildLUT(my_lut);
					aVolume->LoadUserLUT(my_lut);
					break;
				case APROF:
					*AprofMidpoint += (GLint)(25.0*(aWand->RefLoc(0) - aWand->RefOldloc(0)));
					if (*AprofMidpoint < 0)
						*AprofMidpoint = 0;
					if (*AprofMidpoint > 255)
						*AprofMidpoint = 255;
					BuildProf(Aprof, *AprofMidpoint, *AprofHalfwidth, *AprofAmpFactor);
					BuildLUT(my_lut);
					aVolume->LoadUserLUT(my_lut);
					break;
				case ALLPROFS:
					break;
				}
				break;
			case PROF_WIDTH:
				switch (*mlev3_sw) {
				case RPROF:
					*RprofHalfwidth += (GLint)(25.0*(aWand->RefLoc(0) - aWand->RefOldloc(0)));
					if (*RprofHalfwidth < 10)
						*RprofHalfwidth = 10;
					if (*RprofHalfwidth > 500)
						*RprofHalfwidth = 500;
					BuildProf(Rprof, *RprofMidpoint, *RprofHalfwidth, *RprofAmpFactor);
					BuildLUT(my_lut);
					aVolume->LoadUserLUT(my_lut);
					break;
				case GPROF:
					*GprofHalfwidth += (GLint)(25.0*(aWand->RefLoc(0) - aWand->RefOldloc(0)));
					if (*GprofHalfwidth < 10)
						*GprofHalfwidth = 10;
					if (*GprofHalfwidth > 500)
						*GprofHalfwidth = 500;
					BuildProf(Gprof, *GprofMidpoint, *GprofHalfwidth, *GprofAmpFactor);
					BuildLUT(my_lut);
					aVolume->LoadUserLUT(my_lut);
					break;
				case BPROF:
					*BprofHalfwidth += (GLint)(25.0*(aWand->RefLoc(0) - aWand->RefOldloc(0)));
					if (*BprofHalfwidth < 10)
						*BprofHalfwidth = 10;
					if (*BprofHalfwidth > 500)
						*BprofHalfwidth = 500;
					BuildProf(Bprof, *BprofMidpoint, *BprofHalfwidth, *BprofAmpFactor);
					BuildLUT(my_lut);
					aVolume->LoadUserLUT(my_lut);
					break;
				case APROF:
					*AprofHalfwidth += (GLint)(25.0*(aWand->RefLoc(0) - aWand->RefOldloc(0)));
					if (*AprofHalfwidth < 10)
						*AprofHalfwidth = 10;
					if (*AprofHalfwidth > 500)
						*AprofHalfwidth = 500;
					BuildProf(Aprof, *AprofMidpoint, *AprofHalfwidth, *AprofAmpFactor);
					BuildLUT(my_lut);
					aVolume->LoadUserLUT(my_lut);
					break;
				case ALLPROFS:
					break;
				}
				break;
			case PROF_HEIGHT:
				switch (*mlev3_sw) {
				case RPROF:
					*RprofAmpFactor += aWand->RefLoc(0) - aWand->RefOldloc(0);
					if (*RprofAmpFactor < 0.05)
						*RprofAmpFactor = 0.05;
					if (*RprofAmpFactor > 0.5)	//	assures 0.0 < curve <= 1.0
						*RprofAmpFactor = 0.5;
					BuildProf(Rprof, *RprofMidpoint, *RprofHalfwidth, *RprofAmpFactor);
					BuildLUT(my_lut);
					aVolume->LoadUserLUT(my_lut);
					break;
				case GPROF:
					*GprofAmpFactor += aWand->RefLoc(0) - aWand->RefOldloc(0);
					if (*GprofAmpFactor < 0.05)
						*GprofAmpFactor = 0.05;
					if (*GprofAmpFactor > 0.5)	//	assures 0.0 < curve <= 1.0
						*GprofAmpFactor = 0.5;
					BuildProf(Gprof, *GprofMidpoint, *GprofHalfwidth, *GprofAmpFactor);
					BuildLUT(my_lut);
					aVolume->LoadUserLUT(my_lut);
					break;
				case BPROF:
					*BprofAmpFactor += aWand->RefLoc(0) - aWand->RefOldloc(0);
					if (*BprofAmpFactor < 0.05)
						*BprofAmpFactor = 0.05;
					if (*BprofAmpFactor > 0.5)	//	assures 0.0 < curve <= 1.0
						*BprofAmpFactor = 0.5;
					BuildProf(Bprof, *BprofMidpoint, *BprofHalfwidth, *BprofAmpFactor);
					BuildLUT(my_lut);
					aVolume->LoadUserLUT(my_lut);
					break;
				case APROF:
					*AprofAmpFactor += aWand->RefLoc(0) - aWand->RefOldloc(0);
					if (*AprofAmpFactor < 0.05)
						*AprofAmpFactor = 0.05;
					if (*AprofAmpFactor > 0.5)	//	assures 0.0 < curve <= 1.0
						*AprofAmpFactor = 0.5;
					BuildProf(Aprof, *AprofMidpoint, *AprofHalfwidth, *AprofAmpFactor);
					BuildLUT(my_lut);
					aVolume->LoadUserLUT(my_lut);
					break;
				case ALLPROFS:
					break;
				}
				break;
			}
			break;
		case NUM_SLICES:
			*NumPolys += 10.0*(aWand->RefLoc(0) - aWand->RefOldloc(0));
			if (*NumPolys < 1)
				*NumPolys = 1;
			else if (*NumPolys > 500)
				*NumPolys = 500;
//			aVolume->SetNumPolys(*NumPolys);
			break;
		}
	}
	if (!aWand->RefButton2()) {
		RHB2_was_off = 1;
//		*RotAngle = 0.0;
	}
}

void CalcRotParms(Wand *aWand) {
	GLfloat starttemp[3], startvec[3], endtemp[3], endvec[3];
	GLfloat radians;

	//	old vec from data vol center to wand
	starttemp[0] = aWand->RefOldloc(0) - Translate[0];
	starttemp[1] = aWand->RefOldloc(1) - Translate[1];
	starttemp[2] = aWand->RefOldloc(2) - Translate[2];

	//	old vec from data vol center to wand
	endtemp[0] = aWand->RefLoc(0) - Translate[0];
	endtemp[1] = aWand->RefLoc(1) - Translate[1];
	endtemp[2] = aWand->RefLoc(2) - Translate[2];

	//	unitize
	normalize(starttemp, startvec);
	normalize(endtemp, endvec);

	// Calc axis
	crossproduct(RotAxis,startvec,endvec);
	radians = acosf((dot3d(startvec,endvec)));
	*RotAngle = Degrees(radians);
	if (fabsf(*RotAngle) < 1.0) {
		RotAxis[0] = RotAxis[2] = 0.0;
		RotAxis[1] = 1.0;
		*RotAngle = 0.0;
	}
}

GLfloat Degrees(GLfloat radians) {
	return radians * (180.0/M_PI);
}

void crossproduct(float *norm, float *firstvec, float *secondvec) {
	/* The cross product from 1st to 2nd yields the (non-unit) normal */

	norm[0] = (firstvec[1]*secondvec[2] - firstvec[2]*secondvec[1]);
	norm[1] = (firstvec[2]*secondvec[0] - firstvec[0]*secondvec[2]);
	norm[2] = (firstvec[0]*secondvec[1] - firstvec[1]*secondvec[0]);
}

//	Calc's. the 3D dot product FROM 'vec0' TO 'vec1'
float dot3d(float *vec0, float *vec1) {
	return((vec0[0]*vec1[0]) + (vec0[1]*vec1[1]) + (vec0[2]*vec1[2]));
}

//	normalize a 3D vector
void normalize(float *vec, float *unitvec) {
	float magnitude;

	magnitude = (float)pow((double)(vec[0]*vec[0] + vec[1]*vec[1] + vec[2]*vec[2]),(double)0.5);
	unitvec[0] = vec[0]/magnitude;
	unitvec[1] = vec[1]/magnitude;
	unitvec[2] = vec[2]/magnitude;
}

void UpdateAccumMtx() {
}

void ReadFile(char *filename, size_t filesize, void *data) {
	FILE *infile;
	uchar_t *uchar_ptr;
	ushort_t *ushort_ptr;

	uchar_ptr = (uchar_t *)data;
	if ( (infile = fopen (filename, "r")) == (FILE *) NULL ) {
		printf("\n Can't open %s", filename);
		exit(1);
	}
	if (fread((char*)uchar_ptr, sizeof(uchar_t), filesize, infile) != filesize) {
		cout << "Didn't read correct # items from " << filename << endl;
		exit(1);
	}
	fclose(infile);
}

void BuildProf(GLfloat *prof, GLint profMidpoint, GLint profHalfwidth, GLfloat profAmpFactor) {
	//	We consider the cos curve to extend from -PI to PI in all cases;
	//	we only change the mapping of this curve to the indices of the profile.
	//	Initially, the curve is mapped exactly to [0,255].
	//	We can use the wand to narrow or widen the range of indices the curve is mapped to;
	//	we can also use it to move the peek of the curve left and right in order to shift
	//	the entire mapping, however narrow or wide to a lower or higher range of\ profile indices.
	//	We end up with a range of indices, [lomapidx,himapidx], to which the curve range, [-PI,PI] is currently mapped.
	//	Any parts of the curve that are mapped to indices outside the range [0,255] are
	//	clipped.  So the curve may be mapped to any of the 4 following combinations of
	//	profile indices [startidx,endidx]:  [0,255], [lomapidx,himapidx], [0,himapidx], [lomapidx,255].

	GLint	i, temp, lomapidx, himapidx, startidx, endidx;
	GLint lo_outside_buckets, hi_outside_buckets;	//	# indices outside, on both ends, of [0,255]
	GLfloat *f_ptr, Theta;
	GLfloat DTheta;	//	increment in theta per single profile index
	GLfloat starttheta, endtheta;	//	first and last theta values if not entire curve is in [0,255]
	GLboolean DoLoCurve;	//	there is at least 1 bucket, so compute the increasing half of the cos curve
	GLboolean DoHiCurve;	//	there is at least 1 bucket, so compute the decreasing half of the cos curve

	DTheta = M_PI/(float)profHalfwidth;

	//	===>  Right now, I'm going to assume profMidpoint is clamped to [0,255]  <===

	lomapidx = (profMidpoint+1)-profHalfwidth;
	himapidx = profMidpoint + profHalfwidth;
	f_ptr = prof;

	//	For increasing half of cos curve
	if (lomapidx < 0) {
		lo_outside_buckets = 0 - lomapidx;	//	number of Dtheta's left of zero
		if (lo_outside_buckets < profHalfwidth) {	//	there is at least 1 piece of the increasing half of the cos curve
			starttheta = -M_PI+((float)lo_outside_buckets*DTheta);
			startidx = 0;
			DoLoCurve = GL_TRUE;
		} else
			DoLoCurve = GL_FALSE;
	} else {
		startidx = lomapidx;
		starttheta = -M_PI;
		DoLoCurve = GL_TRUE;
	}
	if (DoLoCurve) {
		if (startidx > 0)
			for (i=0; i<startidx; i++) {
				*f_ptr = 0.0;
				f_ptr++;
			}
		Theta = starttheta;
		for (i=startidx; i<profMidpoint; i++) {
			*f_ptr = (cosf(Theta)*profAmpFactor) + profAmpFactor;
			f_ptr++;
			Theta += DTheta;
		}
	}

	//	For decreasing half of cos curve
	if (himapidx > 255) {
		hi_outside_buckets = himapidx - 255;	//	number of Dtheta's right of 255
		if (hi_outside_buckets < profHalfwidth) {
			endtheta = M_PI-(hi_outside_buckets*DTheta);
			endidx = 255;
			DoHiCurve = GL_TRUE;
		} else
			DoHiCurve = GL_FALSE;
	} else {
		endidx = himapidx;
		DoHiCurve = GL_TRUE;
		endtheta = M_PI;
	}
	if (DoHiCurve) {
		Theta = 0.0;
		for (i=profMidpoint; i<=endidx; i++) {
			*f_ptr = (cosf(Theta)*profAmpFactor) + profAmpFactor;
			f_ptr++;
			Theta += DTheta;
		}
		if (endidx < 255)
			for (i=endidx+1; i<256; i++) {
				*f_ptr = 0.0;
				f_ptr++;
			}
	}
}

void BuildLUT (ulong256_t my_lut) {
#if 1
	int i;
	uchar_t *uc_ptr;

	uc_ptr = (uchar_t*)my_lut;
	for (i=0; i<256; i++) {
		*uc_ptr = (uchar_t) (Rprof[i]*256.0);
		uc_ptr++;
		*uc_ptr = (uchar_t) (Gprof[i]*256.0);
		uc_ptr++;
		*uc_ptr = (uchar_t) (Bprof[i]*256.0);
		uc_ptr++;
		*uc_ptr = (uchar_t) (Aprof[i]*256.0);
		uc_ptr++;
	}
#endif
}

void BuildSinLUT(ulong256_t my_lut) {
	int i;
	float two_pi, factor;
	uchar_t *uc_ptr;

	uc_ptr = (uchar_t*)my_lut;
	two_pi = 6.28318;
	//	map two_pi range to 0-255 range
	factor = two_pi/255.0;
	for (i=0; i<256; i++) {
		*uc_ptr = (uchar_t) ((fabs(sin((float)i * factor)))*256.0);
		uc_ptr++;
		*uc_ptr = (uchar_t) ((fabs(sin((float)i * factor)))*256.0);
		uc_ptr++;
		*uc_ptr = (uchar_t) ((fabs(sin((float)i * factor)))*256.0);
		uc_ptr++;
		*uc_ptr = (uchar_t) ((fabs(sin((float)i * factor)))*256.0);
		uc_ptr++;
	}
}

void BuildCosLUT(ulong256_t my_lut) {
	int i;
	float two_pi, factor;
	uchar_t *uc_ptr;

	uc_ptr = (uchar_t*)my_lut;
	two_pi = 6.28318;
	//	map two_pi range to 0-255 range
	factor = two_pi/255.0;
	for (i=0; i<256; i++) {
		*uc_ptr = (uchar_t) ((fabs(cos((float)i * factor)))*256.0);
		uc_ptr++;
		*uc_ptr = (uchar_t) ((fabs(cos((float)i * factor)))*256.0);
		uc_ptr++;
		*uc_ptr = (uchar_t) ((fabs(cos((float)i * factor)))*256.0);
		uc_ptr++;
		*uc_ptr = (uchar_t) ((fabs(cos((float)i * factor)))*256.0);
		uc_ptr++;
	}
}

void BuildJeff2LUT(ulong256_t my_lut) {
	int i;
	float two_pi, factor;
	uchar_t *uc_ptr;

	uc_ptr = (uchar_t*)my_lut;
	two_pi = 6.28318;
	//	map two_pi range to 0-255 range
//	factor = two_pi/255.0;
	factor = 25.13272/255.0;
	for (i=0; i<256; i++) {
		if (i<150) {
			*uc_ptr = (uchar_t) ((fabs(cos((float)i * factor)))*156.0);
			uc_ptr++;
			*uc_ptr = (uchar_t) ((fabs(cos((float)i * factor)))*156.0);
			uc_ptr++;
			*uc_ptr = (uchar_t) ((fabs(cos((float)i * factor)))*0.0);
			uc_ptr++;
			*uc_ptr = (uchar_t) ((fabs(cos((float)i * factor)))*156.0);
			uc_ptr++;
		} else {
			*uc_ptr = (uchar_t) ((fabs(cos((float)i * factor)))*0.0);
			uc_ptr++;
			*uc_ptr = (uchar_t) ((fabs(cos((float)i * factor)))*0.0);
			uc_ptr++;
			*uc_ptr = (uchar_t) ((fabs(cos((float)i * factor)))*156.0);
			uc_ptr++;
			*uc_ptr = (uchar_t) ((fabs(cos((float)i * factor)))*156.0);
			uc_ptr++;
		}
	}
}

void BuildJeffsLUT(ulong256_t my_lut) {
	int i;
	float two_pi, factor, incr, gamma, inv_gamma;
	uchar_t *uc_ptr;

	uc_ptr = (uchar_t*)my_lut;
	incr = 1.0/205.0;
	gamma = 0.6;
	inv_gamma = 1.0/gamma;

	two_pi = 6.28318;
	//	map two_pi range to 0-255 range
	factor = two_pi/255.0;
	for (i=0; i<256; i++) {
		if (i<50) {
			*uc_ptr = (uchar_t)0;
			uc_ptr++;
			*uc_ptr = (uchar_t)0;
			uc_ptr++;
			*uc_ptr = (uchar_t)0;
			uc_ptr++;
			*uc_ptr = (uchar_t)0;
			uc_ptr++;
		} else {
			*uc_ptr = (uchar_t) ((powf(((GLfloat)(i-50)*incr), inv_gamma))*256.0);
			uc_ptr++;
			*uc_ptr = (uchar_t) ((powf(((GLfloat)(i-50)*incr), inv_gamma))*256.0);
			uc_ptr++;
			*uc_ptr = (uchar_t) ((powf(((GLfloat)(i-50)*incr), inv_gamma))*256.0);
			uc_ptr++;
			*uc_ptr = (uchar_t) ((powf(((GLfloat)(i-50)*incr), inv_gamma))*256.0);
			uc_ptr++;
		}
	}
}

void PrintOGLMtx(GLfloat *tempmtx) {
	cout << "     " << tempmtx[0] << "  " << tempmtx[1] << "  " << tempmtx[2] << "  " << tempmtx[3] << endl;
	cout << "     " << tempmtx[4] << "  " << tempmtx[5] << "  " << tempmtx[6] << "  " << tempmtx[7] << endl;
	cout << "     " << tempmtx[8] << "  " << tempmtx[9] << "  " << tempmtx[10] << "  " << tempmtx[11] << endl;
	cout << "     " << tempmtx[12] << "  " << tempmtx[13] << "  " << tempmtx[14] << "  " << tempmtx[15] << endl;
}


#if 0	//	probably don't need this anywhere, don't want to re-type it..
	switch(CurrProf) {
	case RedProf:
		uc_ptr = (uchar_t*)prof;
		break;
	case GreenProf:
		uc_ptr = (uchar_t*)prof;
		uc_ptr += 1;
		break;
	case BlueProf:
		uc_ptr = (uchar_t*)prof;
		uc_ptr += 2;
		break;
	case AlphaProf:
		uc_ptr = (uchar_t*)prof;
		uc_ptr += 2;
		break;
	}
#endif

void DrawProfile(GLfloat *prof, ProfType CurrProf) {
	float vtx[3];
	static int incr = 10;
	int i;

	switch(CurrProf) {
	case RedProf:
		glColor3f(1.0, 0.0, 0.0);
		break;
	case GreenProf:
		glColor3f(0.0, 1.0, 0.0);
		break;
	case BlueProf:
		glColor3f(0.0, 0.0, 1.0);
		break;
	case AlphaProf:
		glColor3f(1.0, 1.0, 1.0);
		break;
	}
	glLineWidth(3.0);
	glPushMatrix();
	glTranslatef(-2.0, 1.0, -5.0);
	glScalef(2.0, 1.0, 1.0);
	glBegin(GL_LINE_STRIP);
	for (i=0; i<256; i+=5) {
		glVertex3f((float)i*.00392, prof[i], 0.0);
	}
	glEnd();
	glPopMatrix();
	glLineWidth(1.0);
}









