//======
//======	gtext.h
#include <stdio.h>
#include <stdarg.h>
#include <cave_ogl.h>
#include "GUIBase.h"

class tGraphicText;
class tSharedMem;
class dataStore;
class clearCom;
class textVec;
class textColor;
class textInfo;

//======	Constant Definitions
//======
#define TX_RIGHT	0
#define TX_CENTER	1
#define TX_LEFT		2

#define TX_CAVE		0
#define TX_WORLD	1

//======	Main Program interface and tGraphicText class
//======
int gprintf (char *format, ...);
int gputs (char *string);

class tGraphicText : public tGUIBase
	{
	friend int gputs (char *string);
	friend dataStore;
	friend clearCom;
  public:
						tGraphicText		(void)
							: tGUIBase (GUI_USE_CAVE | GUI_USE_WORLD ) {} ;
	virtual tGUIMem*	GUISharedMemInit	(void);			// Initiate Shared Memory before processor fork
	virtual void		GUIGraphicsInit		(void);			// Initiate Graphics Memory in CAVEAppInit
	virtual	void		GUIframe			(int);			// Syncronize text output at end ov CAVEFrameFunc

	void clear			(void);										// Clear Screen
	void setHeight		(float);									// Set Height of Letters
	void setDepth		(float);									// Set Depth of Letters
	void setLoc			(const textVec &loc,						// Set Location in CAVE/World
						 const textVec &dirUp,
						 const textVec &dirLeft);
	void setLoc			(float x,     float y,     float z,
						 float xUp,	  float yUp,   float zUp,
						 float xLeft, float yLeft, float zLeft);
	void setColorFront	(textColor color);							// Set Font Color
	void setColorFront	(float x, float y, float z);
	void setColorBody	(textColor color);
	void setColorBody	(float x, float y, float z);
	void setColorBack	(textColor color);
	void setColorBack	(float x, float y, float z);
	void setJustified	(int j);									// Set Text Justification
	void setCoordSystem (int cs);									// Set CAVE/WORLD
	void set			(textInfo &ti);								// Set all Paramiters.

  private:
	void*				sharedMem;		// Pointer to Shared Memory
	void*				textCAVE;		// node containing all of the text
	void*				textWorld;
	};

//======	clearCom - This is a command that clears the Text
//======
class clearCom : public tGUIQue
	{
  public:
	clearCom (tGraphicText *_GT) {GT = _GT;}
	void			doCommand (void);
	tGraphicText	*GT;
	};

//======	textVec - contains three floats.
//======	
struct textVec
	{
	float data[3];
	
	textVec (void)
		{
		data[0] = 1.0;
		data[1] = 0.0;
		data[2] = 0.0;
		}
	textVec (float x, float y, float z)
		{
		data[0] = x;
		data[1] = y;
		data[2] = z;
		}
	};
	
//======	textColor - controls text Color information
//======
struct textColor
	{
	float data[3];
	
	textColor (void)
		{
		data[0] = 1.0;
		data[1] = 0.0;
		data[2] = 0.0;
		}
	textColor (float red, float green, float blue)
		{
		data[0] = red;
		data[1] = green;
		data[2] = blue;
		}
	};
	
//======	textInfo - can to used to set Text Properities.
//======
struct textInfo
	{
	float			height;
	float			depth;
	
	textVec			dirUp,
					dirLeft;
	textVec			loc;
	textColor		front,
					body,
					back;
	int				justified;
	int				coordSystem;
	};	

//======	gtext - Interface to class
//======
extern tGraphicText gtext;
