// $Header: /afs/fl.mcs.anl.gov/usr/cave/hudson/Liaison/ANLE/Frames/OpenGL/inc/RCS/Frame.h,v 1.7 1995/09/27 03:12:08 hudson Exp $

// Reference frame class; not really needed at all unless there are more than 1 frame in the area of interest (a virtual
//                                                                                                     /|\
//                                                                                                      |
// world, a program, an entire application).  If there are more than 1, you, the interested party, -----+
// know that there are, and this usually forces them to be related.  Any two of them are related by the transformation
// (matrix) that takes one of them to the other, and its inverse.  Let any two frames be represented by two nodes in
// a graph, and the matrix and its inverse by edges between the nodes.  Since a matrix is, in general, different from its
// inverse, the edge that represents it must be directed -- and the edge representing its inverse must be directed in
// the opposite direction.  So, one sensible of model a set of frames in a context, and their matrices of relation is
// a digraph.
//       THIS MAY ALL BE TRUE, BUT I'M CURRENTLY THINKING OF MINIMIZING THE # OF LINKS NEEDED IN A GIVEN FRAME TO ITS
//    NEIGHBORS.  SO, I'M THINKING OF JUST HAVING EACH FRAME CONTAIN THE TRANSFORM FROM ITS PARENT TO IT, AND A POINTER
//    TO ITS PARENT.  EACH FRAME COULD (BUT WOULDN'T NEED TO) CONTAIN A LIST OF OBJECTS THAT CLAIM IT AS THEIR FRAME
//    OF REFERENCE.  SO, IF THERE'S AN OBJECT AT THE END OF A CHAIN OF FRAMES, THE OBJECT'S DRAW ROUTINE WOULD APPLY
//    THE TRANSFORMS OF THE FRAMES IN THE CHAIN STARTING WITH ITS OWN FRAME AND WORKING ITS WAY BACK UP THE CHAIN TO
//    THE FRAME THAT IS THE CHILD OF THE (IMPLICIT) WORLD FRAME.  THE WORLD FRAME IS IMPLICIT BECAUSE IT'S NOT
//    INSTANTIATED.
//         |            |          |
//        / \ <------- / \ <----  / \
//     FRAME n+1     FRAME n     WORLD
//
//         |
//         |
//
//         O
//
//       OBJ m
//
//       HOWEVER, THE NON-EXISTENCE OF THE WORLD FRAME, AND THE LACK OF POINTERS BACK DOWN THE FRAME TREE SEEMS TO CREATE
//    A PROBLEM IF I WANT TO USE THE FRAMES TO DO MORE THAN JUST DETERMINE WHERE OBJECTS ARE DRAWN.  WHAT IF I WANT TO
//    USE THE ROBINETT/HOLLOWAY TECHNIQUE OF LOCKING THE HAND TO AN OBJECT AND CALC'ING THE VIRTUAL TRANSFORM BTWN
//    THE HAND AND THE OBJECT?  IT SEEMS I NEED TO MOVE BOTH WAYS IN THE TREE.  IN FACT, EVEN IF I WANT TO MOVE AN
//    OBJECT'S FRAME TO BE THE CHILD OF ANOTHER FRAME IN THE ENVIRONMENT, I'D NEED TO CALC THE VIRTUAL, CURRENT
//    TRANSFORM BTWN THE OBJECT'S FRAME AND ITS NEW PARENT SO I COULD USE THIS TRANSFORM AS THE NEW TRANSFORM-TO-PARENT

// In a computer graphics context, there will be a world frame.  This is sort of the canonical frame in such a context.
// One has to be chosen as the one which will be transformed finally to screen space.  Many of the other frames will
// (sensibly) be related to this one by a single (1-edged) path.  But, many will also be related to the world frame by
// by paths that go through other frames on their way to the world frame.  When using the matrix stack to draw all that is
// in frame X, the composition of matrices to put on the stack is the one represented by the path from the world frame
// to frame X (or, maybe the inverse path).  The world frame doesn't need to be transformed to the world before it is
// drawn, so it doesn't need a transformation matrix.  So, in the general frame which DOES, let the matrix that appears
// there represent the transformation from world-ward sibling frame of the frame in question.  So, if we have
//
//       o <-------- o
//    OBJECT       WORLD
//
// the matrix we find in OBJECT's node represents the edge in the figure.


#pragma once


#include <GL/gl.h>
#include "Shape.h"
// #include "VQS.h"

class Frame {

protected:
	Shape *Geometry;  // `Dummy' ptr -- in specific Frame derived classes, this will be assigned a ptr to specific geometries for those
	                  //	classes
	                  // Model of the frame itself -- axes, usually, I suppose
//	Frame *parentFrame;
	GLboolean visible;
	GLboolean active;
// int dimension; // To be used later; for now, 3d
// int coordtype; // Cartesian, polar, spherical; also to be used later; Cartesian for now
// int spacetype; // Euclidean, hyperbolic, ...
	GLdouble transform[16]; // 4x4 homogeneous matrix containing transformation from this frame's parent to this frame
	GLuint *transform_idx;	//	Index to DispList that applies transform
	GLuint *draw_idx;			//	Index to DispList that draws frame geometry
// VQS *aVQS;  // Vector, quaternion, scalar -- an alternate model of the inter-frame transformation

public:
	Frame();
	~Frame();
	void Draw();
	void InitTransform();		//	Load transform w/ identity matrix
	void GraphicsInit();			//	Init DispList stuff; must be done after OpenGL started
	GLboolean& SetVisible();
	GLboolean& SetActive();
	GLdouble *&SetTransform();	//	Load transform w/ matrix from stack
	GLdouble* GetTransform();	//	Return transform
	void ApplyTransform();		//	Execute transform DispList
	void CplTransform();			//	Compile new transform DispList (w/ same index, however)
//	void SetParentFrame(Frame *ptr);
//	Frame *&SetParentFrame();
//	Frame *GetParentFrame();
};












