// $Header: /afs/fl.mcs.anl.gov/usr/cave/hudson/Liaison/ANLE/Frames/OpenGL/inc/RCS/FramedObj.h,v 1.12 1996/06/24 20:05:44 hudson Exp $

// How to implement the association of an object w/ its reference frame?
// Well, I define a FramedObj class.  All it contains is a pointer to its Frame,
// which, in any application, would be assigned a pointer to a specific type of
// Frame, such as a Cart3DFrame,
// and a void pointer to the actual object, which can be any kind of class.
// To place a FramedObj in another reference frame, assign the pointer to that frame
//	to `theFrame'.


#pragma once

#if !defined(CHILD_ARRAY)
#define CHILD_ARRAY
#endif

#include <GL/gl.h>
#include "Frame.h"
#include "Obj.h"
#if defined(CHILD_LIST_PTR)
#include "list.h"
#elif defined(CHILD_LIST)
#include "list.h"
#endif

const int NUM_CHILDREN = 10;

class FramedObj {

protected:
	Frame *theFrame;
//	GLuint frame_list;	//	Index to CallList that applies frame xform & draws frame geom
	Obj *theObject;
//	GLuint obj_list;		//	Index to CallList that draws object
	FramedObj *parent;
#if defined(CHILD_LIST_PTR)
	LIST<FramedObj> *children;
#elif defined(CHILD_LIST)
	LIST<FramedObj> children;
#elif defined(CHILD_ARRAY)
	GLint num_children;
	FramedObj *children[NUM_CHILDREN];	//	Start out allowing ptr's. to NUM_CHILDREN child FramedObj's
#endif

public:
	FramedObj(Obj *anobj, Frame *aframe);
	~FramedObj();
	void Draw();
	void GraphicsInit();
	Obj *GetTheObject();
	Obj *&SetTheObject();
	Frame *GetTheFrame();
	Frame *&SetTheFrame();
#if defined(CHILD_LIST_PTR)
	LIST<FramedObj> *GetChildren();
	LIST<FramedObj> *&SetChildren();
#elif defined(CHILD_LIST)
	LIST<FramedObj> GetChildren();
	LIST<FramedObj> &SetChildren();
#elif defined(CHILD_ARRAY)
	FramedObj *GetChildren(int i);
	FramedObj *&SetNextChild(FramedObj *child);
#endif
	FramedObj *&SetParent();
	FramedObj *GetParent();
	void ToggleFrameVis();
	void ToggleObjectVis();
};
