//	$Header: /afs/fl.mcs.anl.gov/usr/cave/hudson/C++/OpenGL/include/RCS/Texture3d.h,v 1.10 1996/06/24 22:07:54 hudson Exp $

//	09/27/95:  I had to remind myself that `Texture3d' is a kind of MODEL, not just, stand-alone, a texture


/*	THE FEEDBACK BUFFER CODE WAS BASED ON CODE WRITTEN AT SGI, INC.,
 *	A` LA THE COPYRIGHT NOTICE BELOW.... -- R. Hudson
 *
 * (c) Copyright 1993, Silicon Graphics, Inc.
 * ALL RIGHTS RESERVED 
 */


#pragma once


#if defined(MY_EXT)	//	Not my extensions; my off-ONYX copy of the patch154 extensions
#include "GL/gl.h"
#else
#include <GL/gl.h>
#endif
#include <math.h>
#include "Shape.h"
#include "Model.h"
#include "Reg3dScalField.h"
#include "BBox.h"
#include "types.h"
#include "Utils.h"
#include "FramedObj.h"


//	===============================================================================
//	Child of Data only for size & storage
class FBuffer : public Obj, public Data {
protected:
	GLfloat near, far;
	GLfloat z_viewpt;
	GLfloat data_starts[3];	//	low x,y,z val's of data vol.
	GLfloat data_ends[3];	//	high x,y,z val's of data vol.
	GLfloat data_xtents[3];
	GLfloat texture_starts[3];
	GLfloat texture_ends[3];
	GLfloat texture_xtents[3];
	GLfloat cavewin_starts[3];	//	3d screen coords
	GLfloat cavewin_ends[3];
	GLfloat cavewin_xtents[3];

	//               CAVEWIN                                                              
	//     | -----+-----------------> |                                                                          
	//   xmin   currx               xmax                                                  
	//     | ----->                                                                       
	//                                                                            
	//                DATA                                                                 
	//     | -----------+-----------> |                                                                          
	//  bbox-lo        0.0          bbox-hi                                                                  
	//     | ----->                                                                       
	//                                                                            
	//    changing               unknown                                                                 
	//     currwin    - winmin    slice - bbox-lo                                                  
	//  (from fbuffer)             vtx                                                    
	//  ---------------------- = ----------------------                                                            
	//    winmax - winmin          bbox-hi - bbox-lo                                            
	//                                                                            
	//              TEXTURE                                                                   
	//     | -----------------------> |                                                                          
	//    0.0                        1.0                                                
	//     | ----->                                                                       

	//	constants used in mapping cavewin range to data range; 1 pair in @ of x,y,z
	GLfloat win2data_consts[3][2];
	//	constants used in mapping cavewin range to texture range; 1 pair in @ of x,y,z
	GLfloat win2texture_consts[3][2];

	GLint num_vals_per_vtx;	//	# values returned in FB buffer per returned vertex (cf.
	                        //	OpenGL Prog Guide, table 12-1); depends on type of data returned
	GLint num_values;	//	Total # values (for all returned vertices) returned in FB buffer
public:
	FBuffer(GLint num_polys, BBox *aBBox);
	~FBuffer();
	GLint& SetNum_values();
	void Draw();
	void Cpl_DrawList();
	void GraphicsInit();
   void AllocStorage();
	void *GetStorage();
	void LoadStorage(void *data);
	void ParseDrawFBuffer();
	void PrintFBuffer();
	void PrintFBTokens();
	void BuildFBuffer(FramedObj *fo);
private:
	void PrintLoop(GLint &count);
//	void GetWindowSize();
};


//	===============================================================================
class SamplingPoly : public Obj, public Shape {
protected:
	GLint num_vertices;
	GLfloat vertices[6][3];
	GLfloat l_width;
	GLfloat norm[3];
public:
	SamplingPoly();
	~SamplingPoly();
	GLint &SetNum_vertices();
	GLfloat &SetVertices(int i, int j);
	GLfloat GetVertices(int i, int j);
	GLfloat *&SetNorm();
	GLfloat *GetNorm();
	void Draw();
	void Cpl_DrawList();
	void GraphicsInit();
private:
};


//	===============================================================================
class CanonStack : public Obj, public Shape {
protected:
	GLint num_polys;
	GLfloat stack_extent;	//	Extent of stack in direction _|_ component SamplingPoly's
	SamplingPoly *canonSamplingPoly;
public:
	CanonStack(GLfloat extent, GLint polys);
	~CanonStack();
	void Draw();
	SamplingPoly *GetCanonSamplingPoly();
	void Cpl_DrawList();
	void GraphicsInit();
	void SetNum_polys(int aNum_polys);
};


//	===============================================================================
const int MAX_NUM_POLYS = 500;	//	Causes (expected) maximum storage to be allocated for feedbackbuffer
                                 //	at start of app.; then, needn't do realloc-like growing and shrinking
                                 //	of feedback buffer.  Probably won't need more'n 500 polys for awhile....

enum IRampIdx {RAMP_IDX, REV_RAMP_IDX, NUM_RAMPS_TEX}; // Intensity ramps(fwd&rev.)
enum ARampIdx {OP_RAMP_IDX, REV_OP_RAMP_IDX, NUM_OP_RAMPS_TEX}; // Alpha ramps (fwd & rev.)
const int NUM_TLUT_LIST_INDICES = NUM_RAMPS_TEX + 2; 	//	RAMP_IDX (==0), REV_RAMP_IDX(==1), OP_RAMP_IDX(==0 again),
                                                      //	REV_OP_RAMP_IDX(==1 again) and ANDREA_PROF_IDX

const GLint ANDREA_PROF_IDX = 2;											//	Andrea's color scheme
const GLint USER_PROF_IDX = 3;											//	VRen-user-supplied color scheme

const GLint ANDREA_PROF_BRACKET = 10;									//	range of changing color that straddles each ANDREA_PROF_BND[i]
#if 0
const GLint ANDREA_PROF_BND[6] = {0, 31, 82, 133, 184, 255};	//	indices of tlut elems that bound regions of same color
const GLfloat ANDREA_R_AMNTS[5] = {0.0, 0.0, 0.0, 1.0, 1.0};	//	amounts of red in each region bounded by ANDREA_PROF_BND[]'s
const GLfloat ANDREA_G_AMNTS[5] = {0.0, 0.0, 0.5, 0.5, 1.0};	//	amounts of red in each region bounded by ANDREA_PROF_BND[]'s
const GLfloat ANDREA_B_AMNTS[5] = {0.0, 1.0, 1.0, 0.0, 0.0};	//	amounts of red in each region bounded by ANDREA_PROF_BND[]'s
#else
const GLint ANDREA_PROF_BND[6] = {0, 51, 102, 153, 204, 255};	//	indices of tlut elems that bound regions of same color
const GLfloat ANDREA_R_AMNTS[5] = {0.2, 0.2, 0.2, 1.0, 1.0};	//	amounts of red in each region bounded by ANDREA_PROF_BND[]'s
const GLfloat ANDREA_G_AMNTS[5] = {0.2, 0.2, 0.7, 0.7, 1.0};	//	amounts of red in each region bounded by ANDREA_PROF_BND[]'s
const GLfloat ANDREA_B_AMNTS[5] = {0.2, 1.0, 1.0, 0.2, 0.2};	//	amounts of red in each region bounded by ANDREA_PROF_BND[]'s
#endif

const int NUM_RANGES = 3;
const int MAX_BRICKS = 8;
const int NUM_TLUTS = 4;//	Opacity/intensity ramp, reverse opacity/intensity ramp, Andrea-faithful col tbl, & user col tbl
const int NUM_TLUT_DATAS = 2;
const int FLAT_TLUT = 1;	//	indexes to the array of initial tlut data that is all 255 (a` la volren)

//	Object and Frame id's
enum FOIdx {DATA_BBOX, CANON_STACK, FBUFFER0, FBUFFER1, FBUFFER2, FBUFFER3,
				FBUFFER4, FBUFFER5, FBUFFER6, FBUFFER7, DATA_FO, LOCAL, FBACK_FO, INV_BBOX, NUM_FO};


//	===============================================================================
class	Texture3d : public Model {
protected:
	GLfloat TextureRot[16];			//	rotation (to be applied in texture space) of volume about its center:
	                              //	upper left 3x3 submtx of inverse of calling app's mtx
	GLdouble Zmtx[16];				//	for storing the Z axis as a vertex to be transformed by the data bbox xform
	GLboolean normal_brick_render;//	True, if cos of angle btwn xformed bbox normal & vec from bbox ctr to head is > 0.0; else false
	GLfloat xformed_bbox_norm[3];	//	Z axis, after being xformed by curr BBox rots; compared w/ vec from bbox ctr to head to see
                                 //	which order to render bricks in
	float tex_fraction;	//	the part of the 0.0 <= Z <= 1.0 range occupied by 1 sub bounding box 
	int volume_index;		//	passed in from App::InitShareddata; texindex used to be passed in, but needed multiple texindices
	                     //	for multiple bricks per vol dset.  volume_index is quasi 'virtual texture' index, if 'virtual texture'
	                     //	means the sum of all the texture for a single input dataset
	long tex_mem;
	long single_tex_mem;	//	size of largest possible single texture, given RM board (== 0.5 * tex_mem)
	long num_data_elems;	//	# elements in volume dataset
	long num_data_bytes;	//	num_data_elems * elem size 
	int num_bricks;		//	# of bricks into which the current dataset must be split; must be <= MAX_BRICKS
	long num_brick_elems;	//	# elements in 1 brick
	long num_brick_bytes;	//	# bytes in 1 brick
	GLboolean SavedFrameStates[NUM_FO];		//	Visibilities of frames, saved during GL_FEEDBACK for restoration thereafter
	GLboolean SavedObjectStates[NUM_FO];		//	Visibilities of frames, saved during GL_FEEDBACK for restoration thereafter
	GLfloat DataRotAngle;				//	Angle about its center that data volume spins when user grabs and spins it
	GLfloat DataRotAxis[3];				//	Axis about which data volume spins when user grabs and spins it
	GLfloat SliceRotAngle;				//	Angle about its center that slicing stack spins when user moves
	GLfloat SliceRotAxis[3];				//	Axis about which slicing stack spins when user moves

	// These 2 go together; This is `world frame' local to this class; it's used to render objects in GL_RENDER mode
	Frame *localFrame;
	FramedObj *localFramedObj;

	// These 2 go together; This is `world frame' local to this class it's used to render objects in GL_FEEDBACK mode
	Frame *fbackFrame;
	FramedObj *fbackFramedObj;

	FBuffer *anFBuffer;
	FBuffer *someFBuffer[MAX_BRICKS];	//	each elem. is for a subvolume into which the dataset must be split for bricking
	GLint num_polys;						//	Number of polys in stack
	CanonStack *aCanonStack;
	Utils *someUtils;
	float polyscale;			//	scaling of slicing polys
	GLsizei xres;					//	texture map res, not input data res necessarily
	GLsizei yres;
	GLsizei zres;
	float xlatex;	//	X translation of slicing polys
	float xlatey;	//	Y translation of slicing polys
	float xlatez;	//	Z translation of slicing polys
	unsigned char *three_d_bitmap[MAX_BRICKS];
	ulong256_t *tluts;			//	Array of TLUT's
	int tlutidx;			//	Idx to current tlut
	unsigned int *tlutlistindex;	//	index into list of display list indices to be used in TlutDef
	ulong256_t *tlut_datas;	//	Array of TLUT datasets -- to be operated on by `MapRGB', using the ramp
	int *ramplo;			//	(array of) Index, into 256-element greyramp array, of element where ramp starts
	int *ramphi;			//	Index, into 256-element greyramp array, of element where ramp ends
	float256_t *greyramp;//	(array of) Ramp, in TLUT, from lo - hi end of grey glScalef(cf.`Notes',2/23,&`volren' man);
	                     //	ramp is array of scale factors of corresp. TLUT elements
	int *opramplo;			//	(array of) Index, into 256-element opramp array, of element where ramp starts
	int *opramphi;			//	Index, into 256-element opramp array, of element where ramp ends
	float256_t *opramp;	//	(array of) Ramp, in TLUT, from lo - hi end of opacity scale;
	                     //	ramp is array of scale factors of corresp. TLUT elements
	float256_t *Rprof;	//	(array of) Red color profiles
	float256_t *Gprof;	//	(array of) Green color profiles
	float256_t *Bprof;	//	(array of) Blue color profiles
	float256_t *Aprof;	//	(array of) Alpha color profiles
	unsigned int *texindex;
	int tevindex;
	float *texprops;
	float *tevprops;

public:
	Texture3d(Reg3dScalField *aReg3dScalField, size_t bmapsize, int index, int tex_megs,
				 int aNum_polys = 175, ulong256_t lut = NULL);
	~Texture3d();
	int &SetTlutidx();
	void LoadUserLUT(ulong_t *lut);
	void TexDef(int idx);
	void TevDef(long tevindex);
	void TlutDef(unsigned long *table, int idx);
	void Draw();
	void Cpl_DrawList();
	void Init(Wand *aWand, Head *aHead);
	void OncePerFrame();
#if defined(TLUT4)	//	`ABGR' table component
   unsigned long *RefTextable();
#else	//	`AI' table component
   char *RefTextable();
#endif
	void MoveRamplo(int *thisramplo, int *thisramphi, int rampdelta, int rampidx);
	void MoveRamphi(int *thisramplo, int *thisramphi, int rampdelta, int rampidx);
	void ProcessWand(Wand *aWand);
	void ProcessHead(Head *aHead);
	float GetPolyscale();
	GLfloat& SetPolyscale();
	float GetXlatex();
	float GetXlatey();
	float GetXlatez();
	GLfloat*& SetTextureRot();
	GLfloat& SetXlatex();
	GLfloat& SetXlatey();
	GLfloat& SetXlatez();
	void BuildRamp(int *thisramplo, int *thisramphi, float256_t *thisramp);	//	(re)build greyramp[RAMP_IDX]
	void BuildRevRamp(int *thisramplo, int *thisramphi, float256_t *thisramp);	//	(re)build greyramp[REV_RAMP_IDX]
	void MapRGB(ulong_t *indata, ulong_t *outtlut, float *ramp, float *opramp);//	Use greyramp elems to scale corr. tlut_datas[i] elems & load into tluts[j]
	void NewMapRGBA(ulong_t *indata, ulong_t *outtlut, float *rprof, float *gprof, float *bprof, float *aprof);
	void BuildAndreaProfs();
	int GetRamplo(int i);
	int GetRamphi(int i);
	int GetOpRamplo(int i);
	int GetOpRamphi(int i);
	void MapDataToTex(Reg3dScalField *aReg3dScalField);
	void NewMapDataToTex(Reg3dScalField *aReg3dScalField, int idx);
	void BuildSliceStack();		//	Build stack of MAX_SLICE_RES polys; MAX_SLICE_RES || planes slice the texture volume,
	                           //	which is a rectangular solid;
	                           //	the intersection (possibly empty) of a slicing plane and the texture volume
	                           //	will be a polygon -- naything from a triangle to a hexagon;
	                           //	the polygons will be texture-mapped according to where they slice the 3d texture;
	Obj* GetObj(GLint id);
	Frame* GetFrame(GLint id);
	GLfloat Degrees(GLfloat radians);
	GLfloat Radians(GLfloat degrees);
	void BrickDrawOrder(Head *aHead);

private:
	void crossproduct(float *norm, float *firstvec, float *secondvec);
	float dot3d(float *vec0, float *vec1);
	void normalize(float *vec, float *unitvec);
	void CalcDataRotParms(Wand *aWand);
	void CalcSliceRotParms(Head *aHead);
	void InitSliceRotParms(Head *aHead);
	void PreFeedback();
	void PostFeedback();
	void SaveVisStates();
	void PrintOGLMtx(GLdouble *tempmtx);
};


