/*
 * $Id: hvector.h,v 1.1 1995/11/14 23:59:01 ryan Exp $
 *
 * Copyright (c) 1994,1995,  The University of Chicago
 * All rights reserved
 *
 *                       The University of Chicago
 *                        Department of Radiology
 *
 *
 *     Date              : Jan 1995
 *     Author            : Martin Ryan (m-ryan@uchicago.edu)
 *
 *
 * The permission to use, copy, and redistribute all or any part
 * of the included source code is hereby granted, provided that
 * the above copyright notice and the disclaimer appear unmodified
 * in all copies.
 *
 * DISCLAIMER.
 *
 * The software in this package is provided "as is," without any
 * express or implied warranties, including, without limitation,
 * the implied warranties of merchantibility of fitness for a
 * particular purpose. The author and the University of Chicago
 * are not responsible for any direct or indirect damages resulting
 * from use of this code in any form.
 *
 */

#ifndef _HVECTOR_H_
#define _HVECTOR_H_

#include <iostream.h>
#include <math.h>

class hvector { 
protected:
	float v[4];

	friend class phvector;
	friend class qvector;
	friend class hmatrix;
	friend class hproject;
	friend class PList;

public:
	hvector();	// constructor
	hvector(float);
	hvector(const float[]);
	inline hvector(const hvector*);
	inline hvector(const hvector& hv);
	inline hvector(float x, float y, float z, float w=1);
	float* c_vector() { return &v[0]; }
	float& operator[](int i);
//	float& elem(int i) const { return v[i]; }

	inline float magnitude();
	friend float distance(const hvector&,const hvector&);
	friend float angle(const hvector&,const hvector&);

	hvector& operator ~ ();

	hvector& operator () (float,float,float);

	friend hvector  operator - (const hvector&);

	friend hvector	operator + (const hvector&,const hvector&);
	friend hvector	operator - (const hvector&,const hvector&);
	friend hvector	operator * (const hvector&,const hvector&);
	friend hvector	operator / (const hvector&,const hvector&);

	friend float  	operator ^ (const hvector&,const hvector&);
	friend hvector	operator % (const hvector&,const hvector&);

        inline hvector&	operator += (const hvector&);
        inline hvector&	operator -= (const hvector&);
        inline hvector&	operator *= (const hvector&);
        inline hvector&	operator /= (const float&);

        friend hvector  operator * (const hvector&,const hmatrix&);
        friend hvector  operator * (const hmatrix&,const hvector&);
        friend hvector& operator *= (hvector&,const hmatrix&);

        friend int      operator == (const hvector&,const hvector&);
        friend int      operator != (const hvector&,const hvector&);

	friend ostream& operator << (ostream&,hvector&);
	friend istream& operator >> (istream&,hvector&);

};

inline hvector::hvector(const hvector& hv)
{
	v[0] = hv.v[0]; 
	v[1] = hv.v[1]; 
	v[2] = hv.v[2]; 
	v[3] = hv.v[3]; 
}
 
inline hvector::hvector(const hvector* hvp)
{
        v[0] = hvp->v[0];
        v[1] = hvp->v[1];
        v[2] = hvp->v[2];
        v[3] = hvp->v[3];
}

inline hvector::hvector(float x, float y, float z, float w)
{
	v[0] = x;
	v[1] = y;
	v[2] = z;
	v[3] = w;
}

inline hvector	operator - (const hvector& v1)
{ 
    return hvector(-v1.v[0],-v1.v[1],-v1.v[2]);
}

inline hvector operator + (const hvector& v1,const hvector& v2)
{
	return hvector( v1.v[0]+v2.v[0],
			v1.v[1]+v2.v[1],
			v1.v[2]+v2.v[2]);
}

inline hvector operator - (const hvector& v1,const hvector& v2)
{
	return hvector( v1.v[0]-v2.v[0],
			v1.v[1]-v2.v[1],
			v1.v[2]-v2.v[2]);
}

inline hvector operator * (const hvector& v1,const hvector& v2)
{
	return hvector( v1.v[0]*v2.v[0],
			v1.v[1]*v2.v[1],
			v1.v[2]*v2.v[2]);
}
 
inline hvector operator / (const hvector& v1,const hvector& v2)
{
        return hvector( v1.v[0]/v2.v[0],
                        v1.v[1]/v2.v[1],
                        v1.v[2]/v2.v[2]);
}

inline float	operator ^ (const hvector& v1,const hvector& v2)
{
	return   float(  v1.v[0]*v2.v[0] +
		         v1.v[1]*v2.v[1] +
		         v1.v[2]*v2.v[2]);
}

inline hvector	operator % (const hvector& v1,const hvector& v2)
{
	return hvector( v1.v[1]*v2.v[2] - v1.v[2]*v2.v[1],
			v1.v[2]*v2.v[0] - v1.v[0]*v2.v[2],
			v1.v[0]*v2.v[1] - v1.v[1]*v2.v[0],
			0);
}

inline hvector& hvector::operator += (const hvector& v2)
{
	v[0]+=v2.v[0];
	v[1]+=v2.v[1];
	v[2]+=v2.v[2];
	return *this;
}

inline hvector& hvector::operator -= (const hvector& v2)
{
        v[0]-=v2.v[0];
        v[1]-=v2.v[1];
        v[2]-=v2.v[2];
	return *this;
}

inline hvector& hvector::operator *= (const hvector& v2)
{
	v[0]*=v2.v[0];
	v[1]*=v2.v[1];
	v[2]*=v2.v[2];
	return *this;
}

inline hvector& hvector::operator /= (const float& f)
{
	float invf=1/f;
        v[0]*=invf;
        v[1]*=invf;
        v[2]*=invf;
        return *this;
}

inline int operator == (const hvector& hv1,const hvector& hv2)
{
    return hv1.v[0] == hv2.v[0] &&
           hv1.v[1] == hv2.v[1] &&
           hv1.v[2] == hv2.v[2];
}

inline int operator != (const hvector& hv1,const hvector& hv2)
{
    return hv1.v[0] != hv2.v[0] ||
           hv1.v[1] != hv2.v[1] ||
           hv1.v[2] != hv2.v[2];
}

inline hvector& hvector::operator () (float x,float y,float z)
{
    v[0]=x;
    v[1]=y;
    v[2]=z;
    return *this;
}

inline float hvector::magnitude()
{
	return sqrt(*this ^ *this);
}

class phclass {
public:
        enum { comma=01, homog=02, paren=04,
               brace=010, bracket=020 };
protected:
        long  flags;
        int   open(ostream&) const;
        int   close(ostream&) const;
        print_homogeneous() const { return flags&homog; }
        char* commaSet() const { return flags&comma ? "," : " "; }

	phclass(long argflags) { flags=argflags; }
	phclass(const char*);
};

class phvector : public phclass {
	const hvector* hvp;
	const float& operator [] (int i) const { return hvp->v[i]; }
public:
	phvector(const hvector&,long =comma);
	phvector(const hvector&,const char*);
	friend ostream& operator << (ostream&,const phvector&);
};

#endif /* _HVECTOR_H_ */
